#!/usr/bin/env python3

import getpass
import os
from typing import List
import textwrap

from openai import OpenAI
from openai.types.chat.chat_completion_message import ChatCompletionMessage

from ..config import Config
from ..util import unquote_all

class DefaultDispatcher:
  """
  Default fallback dispatcher for llm2sh, used when there isn't a more specific
  dispatcher available. This dispatcher can be extended to customize the
  system prompt or to use a different API.
  """

  def __init__(self, uri: str, key: str, model: str, config: Config, temperature: float, verbose: bool = False):
    self.uri = uri
    self.key = key
    self.model = model
    self.verbose = verbose
    self.config = config
    self.temperature = temperature


  def dispatch(self, request_str: str):
    """
    Sends the request to the LLM, returning the series of commands to run.
    """

    system_prompt = self._get_system_prompt(request_str)
    if self.verbose:
      print(f"[DEBUG]: System prompt:\n{system_prompt}")

    client = OpenAI(
      base_url=None if self.uri == '' else self.uri,
      api_key='NA' if self.key == '' else self.key,
    )
    message = client.chat.completions.create(
        model=self.model,
        temperature=self.temperature,
        messages=[
          { "role": "system", "content": system_prompt },
          { "role": "user", "content": request_str }
        ],
    )

    response = message.choices[0].message.content
    if self.verbose:
      print(f"[DEBUG]: Response:\n{response}")

    return self._clean_output(response.split('\n'))


  def _max_context_length(self) -> int:
    """
    Returns the input maximum length of the LLM model.
    """
    return 4096  # Assume the worst case scenario (i.e. gpt-3.5-turbo-instruct supports a 4K window)


  def _max_system_prompt_length(self, request_str: str) -> int:
    """
    Returns the input maximum length of the system prompt.
    Reserve half the context length for the output, and the remaining amount (less the request length) as
    system prompt length.

    Assume each token is ~4 characters long.
    """
    return (self._max_context_length() // 2) - (len(request_str) // 4) + 1


  def _whoami(self) -> str:
    """
    Returns the current user.
    """
    return getpass.getuser()


  def _cwd(self) -> str:
    """
    Returns the current working directory.
    """
    return os.getcwd()


  def _ls(self, summarize_factor: int) -> List[str]:
    """
    Returns the contents of the current working directory.
    The `summarize_factor` is used to suppress overly long directory listings.
    """
    items = os.listdir()
    return items[:int(len(items) * summarize_factor)]


  def _additional_context(self, summarize_factor: int) -> str:
    """
    Returns additional situational context to be included in the system prompt.
    """

    lines = []

    # If lsb_release is available, include the OS version
    if os.path.exists('/usr/bin/lsb_release'):
      lines.append('The OS is:')
      lines.append(os.popen('/usr/bin/lsb_release -d').read().strip())

    return '\n'.join(lines)


  def _available_env(self, summarize_factor: int) -> List[str]:
    """
    Returns the list of available environment variables.
    The `summarize_factor` is used to limit the number of results.
    This method also omits some common environment variables that are usually not useful.
    """
    items = set(os.environ.keys())

    # Remove common junk environment variables by name. Additional ones are filtered out via patterns.
    omit = {
      # Terminal color settings
      'color_prompt', 'force_color_prompt',
      'COLORTERM', 'LSCOLORS', 'LS_COLORS', 'LS_OPTIONS', 'CLICOLOR',

      # GUI & Auth implementation details
      'SESSION_MANAGER', 'TERM_PROGRAM_VERSION', 'VDPAU_DRIVER',
      'SSH_AUTH_SOCK', 'SYSTEMD_EXEC_PID', 'XAUTHORITY',

      # Misc others
      'MOTD_SHOWN', 'PYTHONSTARTUP', 'INVOCATION_ID'
    }
    items = items - omit
    items = [
      i for i in items
      if (
        # Inserted when running in a VSCode terminal
        (not i.startswith('VSCODE_'))

        # Color codes
        and (not i.startswith('COLOR_'))

        # Misc others
        and (not i.startswith('XDG_'))
        and (not i.startswith('DBUS_'))
        and (not i.startswith('GJS_'))
        and (not i.startswith('GDM_'))
        and (not i.startswith('GIO_'))
      )
    ]

    # Always keep a minimum of 20 items - there's more opportunity to remove items from the
    # directory contents listing than the environment variables listing.
    take_n = min(int(len(items) * summarize_factor), 20)
    return items[:take_n]


  def _get_system_prompt(self, request_str: str) -> str:
    """
    Returns the system prompt.
    """
    max_length = self._max_system_prompt_length(request_str)
    summarize_factor = 1.0  # Percentage of the env and directory information to include in the prompt

    def get_prompt(factor: float) -> str:
      nl = '\n'
      return textwrap.dedent(f"""
        You are an AI helping the user perform tasks in a Bash shell. The user will give you a request,
        and you will generate one or more shell commands to fulfill that request. You can use any shell constructs,
        including pipes, redirection, and loops. You can also use any commands available in the shell environment.
        The shell is configured with `set -e`, so generate appropriate error handling for commands that are allowed
        to fail.

        The user is currently logged in as `{self._whoami()}` and the current working directory is `{self._cwd()}`.

        The current directory contains the following files: \n{nl.join([f"         - {i}" for i in self._ls(factor)])}

        You can refer to the following environment variables: \n{nl.join([f"         - {i}" for i in self._available_env(factor)])}

        \n{self._additional_context(factor)}

        Make sure you output valid shell commands, paying special attention to quoting and escaping.
        Do not wrap the response in quotes or backticks. If you want to provide additional information,
        please include it in a shell comment (i.e. `#`) or use `echo`. For more complex tasks, you can
        use `cat` to write a Python script to a file and then execute it.

        YOU MUST RESPOND WITH ONLY VALID SHELL COMMANDS. DO NOT INCLUDE ANYTHING ELSE IN YOUR RESPONSE.
      """)

    prompt = get_prompt(summarize_factor)
    while len(prompt) // 4 > max_length:
      # Reduce the amount of information in the prompt
      summarize_factor = summarize_factor * 0.9 * (max_length / (len(prompt) // 4))
      prompt = get_prompt(summarize_factor)

    return prompt


  def _clean_output(self, output: List[str]) -> List[str]:
    """
    Cleans the output by removing any leading or trailing whitespace and other common mistakes.
    """

    # Remove trailing whitespace. Keep leading whitespace in case the LLM is trying to
    # generate a Python script or something similar.
    cleaned = [i.rstrip() for i in output]

    cleaned = [
      # Remove quotes around the commands
      unquote_all(i, ['`', '"', "'"])

      for i in cleaned

      if not (
        # Heuristic: Remove empty lines and code blocks
        (i == '')
        or (i == '```')
        or (i == '```bash')
        or (i == '```shell')
        or (i == '```sh')

        # Heuristic: Remove lines where the LLM is trying to be conversational
        or (i.startswith('Sure'))  # i.e. "Sure! Here is..."
        or (i.startswith('Here'))  # i.e. "Here is..."
      )
    ]

    return cleaned



#!/usr/bin/env python3

from typing import List

from anthropic import Anthropic

from .DefaultDispatcher import DefaultDispatcher
from ..config import Config


class AnthropicDispatcher(DefaultDispatcher):
  def __init__(self, key: str, model: str, config: Config, temperature: float, verbose: bool = False):
    super().__init__('', key, model, config, temperature, verbose)


  def dispatch(self, request_str: str):
    """
    Sends the request to the LLM, returning the series of commands to run.
    """

    system_prompt = self._get_system_prompt(request_str)
    if self.verbose:
      print(f"[DEBUG]: System prompt:\n{system_prompt}")

    client = Anthropic(api_key=self.key)
    message = client.messages.create(
        model=self.model,
        max_tokens=2000,
        temperature=self.temperature,
        system=system_prompt,
        messages=[
          { "role": "user", "content": request_str }
        ],
    )

    response = message.content[0].text
    if self.verbose:
      print(f"[DEBUG]: Response:\n{response}")

    return self._clean_output(response.split('\n'))


  def _max_context_length(self) -> int:
    """
    Returns the input maximum length of the LLM model.
    """
    return 16*1024  # Technically larger than this, but this is a reasonable upper bound



#!/usr/bin/env python3

import fcntl
import os
import sys
from typing import List

def instance_already_running(label="default"):
  lock_file_pointer = os.open(f"/tmp/instance_{label}.lock", os.O_WRONLY | os.O_CREAT)

  try:
    fcntl.lockf(lock_file_pointer, fcntl.LOCK_EX | fcntl.LOCK_NB)
    already_running = False
  except IOError:
    already_running = True

  return already_running


def eprint(*args, **kwargs):
  print(*args, file=sys.stderr, **kwargs)


def ethrow(*args, **kwargs):
  eprint(*args, **kwargs)
  raise Exception(*args)


def remove_if_exists(path, *args, **kwargs):
  if os.path.exists(path):
    os.remove(path, *args, **kwargs)


def flatten(xs):
    result = []
    if isinstance(xs, (list, tuple)):
        for x in xs:
            result.extend(flatten(x))
    else:
        result.append(xs)
    return result


def quote(x: str) -> str:
  return f'"{x}"'


def unquote(x: str, quote='"') -> str:
  if x.startswith(quote) and x.endswith(quote):
    return x[1:-1]
  return x


def unquote_all(x: str, quotes: List[str]) -> str:
  for quote in quotes:
    x = unquote(x, quote)
  return x


def nullif(value: str, null_value: str) -> str:
  return None if value == null_value else value



#!/usr/bin/env python3

import argparse
import os
import subprocess

from typing import List, Optional, Dict, Set, Tuple

from .config import Config
from .util import ethrow
from .dispatchers.DefaultDispatcher import DefaultDispatcher
from .dispatchers.AnthropicDispatcher import AnthropicDispatcher

class Cli(object):
  def __init__(self):
    parser = argparse.ArgumentParser(
      description='Ask GPT to run a command',
      usage='llm2sh [options] <request>',
      add_help = True,
    )

    parser.add_argument('request', help = 'request for gpt', action='store', nargs='*')
    # parser.add_argument('--setup', help = 'configure llm2sh', action = 'store_true')
    parser.add_argument('-c', '--config', help = 'specify config file, (Default: ~/.config/llm2sh/llm2sh.json)',
                        action = 'store', default = '~/.config/llm2sh/llm2sh.json')
    parser.add_argument('-d', '--dry-run', help = 'do not run the generated command', action = 'store_true')
    parser.add_argument('-l', '--list-models', help = 'list available models', action = 'store_true')
    parser.add_argument('-m', '--model', help = 'specify which model to use', action = 'store')
    parser.add_argument('-s', '--silent', help = 'don\'t ask bash to output each command before running it.', action = 'store_true')
    parser.add_argument('-t', '--temperature', help = 'use a custom sampling temperature', action = 'store')
    parser.add_argument('-v', '--verbose', help = 'print verbose debug information', action = 'store_true')
    parser.add_argument('-f', '--yolo', '--force', help = 'run whatever GPT wants, without confirmation', action = 'store_true')

    #
    # Load Config
    #
    self.args = parser.parse_args()

    # if self.args.setup:
    #   # self.setup()
    #   ethrow('[ERROR]: Setup not implemented')

    self.config = self.load_config(self.args.config)

    #
    # Load Models
    #
    self.model_list = self.load_models()
    self.models = {i[0]: i for i in self.model_list if i[1]}

    if self.args.list_models:
      self.list_models()
      return

    self.selected_model = self.args.model or self.config.default_model
    if self.selected_model not in self.models:
      ethrow(f"Model {self.selected_model} not available! Configure {self.args.config} or use --list-models to see available models.")

    #
    # Run Request
    #
    if len(self.args.request) == 0:
      parser.print_help()
      return
    self.request_str = ' '.join(self.args.request)
    response = self.dispatch_request(self.request_str)
    self.run_request(response)


  def load_config(self, config_file: str) -> Config:
    config_file = os.path.expanduser(config_file)

    if not os.path.exists(config_file):
      if self.args.verbose:
        print(f"[DEBUG] Config file {config_file} not found, creating a new one.")
      config = Config()
    else:
      config = Config.load_config(config_file)

    # Try to write the config back in case new fields were added.
    config.save_config(config_file)
    return config


  def list_models(self):
    print('Available models:')
    print(f'Models can be configured via {self.args.config}\n')

    max_just_name = max([len(module) for (module, _, _, _, _) in self.model_list])
    max_just_avail = len('NOT AVAILABLE')

    for (model, avail, help, _, model_id) in self.model_list:
      print(
        model.ljust(max_just_name + 2) +
        ('OK' if avail else 'NOT AVAILABLE').rjust(max_just_avail) +
        " | "
        f"{help} ({model_id})"
      )


  def load_models(self) -> List[Tuple[str, bool, str, str]]:
    openai_available = len(self.config.effective_openai_key) > 0
    openai_str = "Ready" if openai_available else f"Requires OpenAI API key"

    claude_available = len(self.config.effective_claude_key) > 0
    claude_str = "Ready" if claude_available else f"Requires Claude API key"

    groq_available = len(self.config.effective_groq_key) > 0
    groq_str = "Ready" if groq_available else f"Requires Groq API key"

    local_available = len(self.config.local_uri) > 0
    local_str = f"Ready - {self.config.local_uri}" if local_available else f"Requires local LLM API URI"

    return [
      ('local', local_available, local_str, 'LOCAL', 'local'),
      ('gpt-4o', openai_available, openai_str, 'OPENAI', 'gpt-4o'),
      ('gpt-3.5-turbo-instruct', openai_available, openai_str, 'OPENAI', 'gpt-3.5-turbo-instruct'),
      ('gpt-4-turbo', openai_available, openai_str, 'OPENAI', 'gpt-4-turbo'),

      ('claude-3-opus', claude_available, claude_str, 'CLAUDE', 'claude-3-opus-20240229'),
      ('claude-3-sonnet', claude_available, claude_str, 'CLAUDE', 'claude-3-sonnet-20240229'),
      ('claude-3-haiku', claude_available, claude_str, 'CLAUDE', 'claude-3-haiku-20240307'),

      ('groq-llama3-8b', groq_available, groq_str, 'GROQ', 'llama3-8b-8192'),
      ('groq-llama3-70b', groq_available, groq_str, 'GROQ', 'llama3-70b-8192'),
      ('groq-mixtral-8x7b', groq_available, groq_str, 'GROQ', 'mixtral-8x7b-32768'),
      ('groq-gemma-7b', groq_available, groq_str, 'GROQ', 'gemma-7b-it'),
    ]


  def dispatch_request(self, request: str) -> List[str]:
    (_, _, _, model_type, model_id) = self.models[self.selected_model]
    temperature = self.args.temperature or self.config.temperature

    # TOTO: Implement dispatcher selection based on the chosen model
    if model_type == 'OPENAI':
      dispatcher = DefaultDispatcher(
        uri = '',  # Defaults to None, which uses the OpenAI API
        key = self.config.effective_openai_key,
        model = model_id,
        config = self.config,
        temperature=temperature,
        verbose=self.args.verbose,
      )
    elif model_type == 'CLAUDE':
      dispatcher = AnthropicDispatcher(
        key = self.config.effective_claude_key,
        model = model_id,
        config = self.config,
        temperature=temperature,
        verbose=self.args.verbose,
      )
    elif model_type == 'GROQ':
      dispatcher = DefaultDispatcher(
        uri = 'https://api.groq.com/openai/v1',
        key = self.config.effective_groq_key,
        model = model_id,
        config = self.config,
        temperature=temperature,
        verbose=self.args.verbose,
      )
    elif model_type == 'LOCAL':
      dispatcher = DefaultDispatcher(
        uri = self.config.local_uri,
        key = self.config.local_api_key,
        model = self.config.local_model_name,
        config = self.config,
        temperature=temperature,
        verbose=self.args.verbose,
      )
    else:
      ethrow(f"Unknown model type {model_type}")

    return dispatcher.dispatch(request)


  def run_request(self, response: List[str]):
    if not self.args.dry_run:
      print('You are about to run the following commands:')
    else:
      print('(Dry Run) The LLM suggested these commands:')

    for command in response:
      print(f'  $ {command}')

    if self.args.dry_run:
      return

    # Prompt the user for confirmation
    if (not self.args.yolo) and (not self.config.i_like_to_live_dangerously):
      print('Run the above commands? [y/N]')
      if input().lower() != 'y':
        print('Request canceled')
        return

    self.run_commands(response)


  def run_commands(self, commands: List[str]):
    # Create a Popen object with shell=True to run commands in a bash shell
    process = subprocess.Popen(
        ["bash"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,  # Redirect stderr to stdout
        text=True,
        shell=True
    )

    # Configure the target shell - abort on errors
    process.stdin.write('set -e\n')
    if not self.args.silent:
      process.stdin.write('set -o xtrace\n')
    process.stdin.flush()

    # Send all the commands at once and terminate the input stream so
    # we don't hang on reading the output
    for command in commands:
      process.stdin.write(command + '\n')
      process.stdin.flush()
    process.stdin.close()

    # Read the output of the command
    while process.stdout.readable():
        output = process.stdout.readline()
        if not output:
            break
        print(output.rstrip())

    # Close the stdin and wait for the process to finish
    process.wait()



import json
import os
from io import StringIO
from dataclasses import dataclass, asdict

@dataclass
class Config:
  default_model: str = 'gpt-4o'

  openai_api_key: str = ''
  claude_api_key: str = ''
  groq_api_key: str = ''

  local_uri: str = 'http://localhost:5000/v1'
  local_api_key: str = ''
  local_model_name: str = ''

  temperature: float = 0.2

  # Always run in yolo mode
  i_like_to_live_dangerously: bool = False


  @staticmethod
  def load_config(config_file: str) -> "Config":
    with open(config_file, 'r') as f:
      return Config.from_dict(json.load(f))


  @classmethod
  def from_dict(cls, d: dict) -> "Config":
    return cls(
      default_model = d.get('default_model', 'gpt-3.5-turbo'),
      openai_api_key = d.get('openai_api_key', ''),
      claude_api_key = d.get('claude_api_key', ''),
      groq_api_key = d.get('groq_api_key', ''),
      local_uri = d.get('local_uri', 'http://localhost:5000/v1'),
      local_api_key = d.get('local_api_key', ''),
      temperature = d.get('temperature', 0.2),
      i_like_to_live_dangerously = d.get('i_like_to_live_dangerously', False),
    )


  @property
  def effective_openai_key(self) -> str:
    if len(self.openai_api_key) > 0:
      return self.openai_api_key
    elif len(os.environ.get('OPENAI_API_KEY', '')) > 0:
      return os.environ['OPENAI_API_KEY']
    else:
      return ''


  @property
  def effective_claude_key(self) -> str:
    if len(self.claude_api_key) > 0:
      return self.claude_api_key
    elif len(os.environ.get('ANTHROPIC_API_KEY', '')) > 0:
      return os.environ['ANTHROPIC_API_KEY']
    else:
      return ''


  @property
  def effective_groq_key(self) -> str:
    if len(self.groq_api_key) > 0:
      return self.groq_api_key
    elif len(os.environ.get('GROQ_API_KEY', '')) > 0:
      return os.environ['GROQ_API_KEY']
    else:
      return ''


  def save_config(self, config_file: str):
    # There are cases where the config location is read-only. Swallow the exception
    try:
      os.makedirs(os.path.dirname(config_file), exist_ok=True)
      with open(config_file, 'w') as f:
        json.dump(asdict(self), f, indent=4)
    except Exception as e:
      pass
